const logger = require("./common").createLogger();
const stringify = require("json-stringify");
const randomitem = require("random-item");
const uniquerandom = require("unique-random");
const randomint = require("random-int");
const SPAWN_DETAILS_KEY = "spwanDetails";
const SPAWN_INDEX = "__spawn_index";
const END_REASON = "Game Had Ended";
const spawnNewFruits = function(game) {
  let spawnDetails = game.getTransient(SPAWN_DETAILS_KEY);
  if (!spawnDetails) return logger.error("spawn details object not found!!");
  if (spawnDetails.nextSpawnTime > Date.now()) return;
  let currentSpawnTime = spawnDetails.nextSpawnTime;
  logger.trace("spawning time....");
  //generate a random array of spawn fruit indexes

  let fruitsToBeSpawned = getSpawnList.call(this, game);
  if (logger.trace()) {
    logger.trace("will spawn fruits %s ", stringify(fruitsToBeSpawned));
  }
  sendSpawnOrder.call(this, game, fruitsToBeSpawned);
  spawnDetails.nextSpawnTime =
    currentSpawnTime + this.meta.extras.spawnFrequency;
};

const getSpawnList = function(game) {
  let spawnPattern = game.getTransient(SPAWN_DETAILS_KEY);
  let spawnIndex = game.getTransient(SPAWN_INDEX);
  logger.trace(
    "Will spawn %s fruits @slot %s",
    spawnPattern.spawnCount,
    spawnIndex
  );
  let fruitsToBeSpawned = [];
  let specialFruits = spawnPattern.spawnLists[spawnIndex];
  specialFruits.forEach(t => {
    fruitsToBeSpawned.push(t);
  });
  for (let c = 0; c < spawnPattern.spawnCount - specialFruits.length; c++) {
    fruitsToBeSpawned.push(randomitem(this.normalFruitTypes));
  }
  game.addTransient(SPAWN_INDEX, ++spawnIndex);
  game.setGameProp(SPAWN_INDEX, spawnIndex.toString());
  return fruitsToBeSpawned;
};

const endGame = function(game) {
  let endTime = game.getScheduledStartTime() + game.getDuration();
  if (endTime < Date.now()) {
    //end game
    logger.debug("Ending fruit ninja instance %s", game.getId());
    game.endGame(END_REASON, err => {
      if (err) {
        logger.error({ err: err });
        logger.error("Error in ending game");
      }
    });
    return true;
  }
  return false;
};
const step = function(game) {
  let gameState = game.getState();
  switch (gameState) {
    case 20: //running
      if (endGame.call(this, game)) return;
      spawnNewFruits.call(this, game);
      break;
    default:
      logger.trace("gamestate is %s, nothing to do ", gameState);
  }
};

const sendSpawnOrder = function(game, fruitsToBeSpawned) {
  let ts = Date.now();
  let connectedPlayers = game.getConnectedPlayers();
  let expiryTime = Date.now() + this.meta.duration;
  connectedPlayers.forEach(player => {
    let gameObjects = [];
    for (let i = 0; i < fruitsToBeSpawned.length; i++) {
      let type = fruitsToBeSpawned[i];
      let gameObject = {
        type: type,
        id: player + "_" + type + "_" + (ts + i),
        owner: player,
        expiryTime: expiryTime
      };
      gameObjects.push(gameObject);
    }
    game.spawnObjects(gameObjects, false, player);
  });
};

const initFruitTypes = function(meta) {
  if (!meta) return;
  //create an object type array
  let fruitTypes = [];
  meta.gameObjects.forEach(obj => {
    if (obj.type === "blade") return;
    fruitTypes.push(obj.type);
  });
  this.fruitTypes = fruitTypes;
  this.normalFruitTypes = meta.extras.normalfruits;
  this.meta = meta;
};

const initSpawnPattern = function(meta) {
  let spawnBounds = meta.extras.spwanBounds;
  let totalSpawnEventCount = -1;
  let maxSpawnUpperBound = -1;
  for (let [type, bound] of Object.entries(spawnBounds)) {
    maxSpawnUpperBound =
      bound[1] > maxSpawnUpperBound ? bound[1] : maxSpawnUpperBound;
  }
  if (maxSpawnUpperBound > -1) {
    //we have some bound
    totalSpawnEventCount = maxSpawnUpperBound + 2; //plus 2 for randomization
  }
  let spawnEventCountStatic = meta.duration / meta.extras.spawnFrequency;
  let calculatedSpawnEventCount =
    spawnEventCountStatic > totalSpawnEventCount
      ? spawnEventCountStatic
      : totalSpawnEventCount;
  let spawnCountPerEvent = Math.ceil(
    (spawnEventCountStatic * meta.extras.spawnCount) / calculatedSpawnEventCount
  );
  let calculatedSpawnFreq = Math.ceil(
    meta.duration / calculatedSpawnEventCount
  );
  //lets create the predefine array
  let spawnLists = [];
  for (let i = 0; i < calculatedSpawnEventCount; i++) {
    let spawnList = [];
    spawnLists.push(spawnList);
  }

  for (let [type, bound] of Object.entries(spawnBounds)) {
    let uRandom = uniquerandom(0, calculatedSpawnEventCount - 1);
    for (let c = 0; c < bound[1]; c++) {
      let toAdd = true;
      if (c > bound[0]) {
        toAdd = randomint(0, 1) == 1;
      }
      if (!toAdd) continue;
      let idx = uRandom();
      spawnLists[idx].push(type);
    }
  }
  let timeBounds = meta.extras.spawnTimebounds;
  for (let [type, bounds] of Object.entries(timeBounds)) {
    for (let i = 0; i < bounds.length; i++) {
      let bound = bounds[i];
      let lowTime = bound[0];
      let hiTime = bound[1];
      let lowSlot = Math.floor(lowTime / calculatedSpawnFreq);
      let hiSlot = Math.floor(hiTime / calculatedSpawnFreq);
      let slotIdx = randomint(lowSlot, hiSlot);
      spawnLists[slotIdx].push(type);
    }
  }
  //create spwanDetails Object
  let spawnPattern = {
    spawnFrequency: calculatedSpawnFreq,
    spawnCount: spawnCountPerEvent,
    spawnLists: spawnLists
  };
  if (logger.debug()) {
    logger.debug("Spawn pattern is %s", stringify(spawnPattern));
  }
  return spawnPattern;
};

const startGame = function(game, cb, meta) {
  let spawnPattern = initSpawnPattern.call(this, meta);
  spawnPattern.nextSpawnTime = Date.now();
  game.addTransient(SPAWN_DETAILS_KEY, spawnPattern);
  game.setGameProp(SPAWN_DETAILS_KEY, JSON.stringify(spawnPattern));
  game.addTransient(SPAWN_INDEX, 0);
  game.setGameProp(SPAWN_INDEX, "0");
  //get all the players
  let playerIds = game.getAllPlayerIds();
  if (!playerIds) return cb(null, null);
  if (playerIds.length < 2) return cb(null, null);
  //assign a knife to each one of them
  let knives = game.gameObjectByIds("knife1", "knife2");
  game.changeOwner(playerIds[0], "_knife1");
  game.changeOwner(playerIds[1], "_knife2");
  cb(null, null);
};

const resumeGame = function(game) {
  logger.info("resuming game ");
  let spawnDetailsString = game.getGameProp(SPAWN_INDEX);
  let spawnDetails = null;
  try {
    spawnDetails = JSON.parse(spawnDetailsString);
  } catch (err) {
    logger.error({ err: err });
    logger.error("Error in parsing spawnPattern String");
  }
  spawnDetails =
    spawnDetails == null
      ? initSpawnPattern.call(this, this.meta)
      : spawnDetails;
  game.addTransient(SPAWN_DETAILS_KEY, spawnDetails);
  //
  let spawnIndex = parseInt(game.getGameProp(SPAWN_INDEX));
  spawnIndex = isNaN(spawnIndex) ? 0 : spawnIndex;
  game.addTransient(SPAWN_INDEX, spawnIndex);
};

const gameLoaded = function(meta) {
  logger.info("loading defs from meta %s", stringify(meta));
  initFruitTypes.call(this, meta);
};

const gameResult = function(game) {
  //get all playerIds
  if (game.getState() !== 30) {
    return [];
  }
  let playerIds = game.getAllPlayerIds();
  if (logger.debug()) {
    logger.debug("playerd who played %s", stringify(playerIds));
  }
  if (playerIds == null || playerIds == undefined || playerIds.length !== 2) {
    logger.error("Fatal error, player count != 2 ");
    throw { err: "INVALID_PLAYER_COUNT" };
  }
  let p1Score = game.getPlayerScore(playerIds[0]);
  let p2Score = game.getPlayerScore(playerIds[1]);
  logger.info(
    "scores for players %s=%s, %s=%s",
    playerIds[0],
    p1Score,
    playerIds[1],
    p2Score
  );
  if (p1Score === p2Score) {
    return [
      { playerId: playerIds[0], winType: 3 },
      { playerId: playerIds[1], winType: 3 }
    ];
  }
  if (p1Score < p2Score) {
    return [
      { playerId: playerIds[0], winType: 2 },
      { playerId: playerIds[1], winType: 1 }
    ];
  }
  return [
    { playerId: playerIds[0], winType: 1 },
    { playerId: playerIds[1], winType: 2 }
  ];
};

module.exports = { startGame, step, gameLoaded, resumeGame, gameResult };
