const logger = require("./common").createLogger();
const stringify = require("json-stringify");
const KNIFE_ID = "_knife";
const BOMB_HIT_KEY = "_bomb_hit";
const COMBO_ARR = "_combo_arr";
const playerDisconnection = function(game, playerId) {
  logger.trace(" player %s  disconnected", playerId);
  delete this.players[playerId];
};

const playerJoined = function(game, playerId) {
  logger.trace("player %s joined ", playerId);
  this.players[playerId] = true;
};

const gameObjectsDestroyed = function(game, gameObjectIds, playerId) {
  if (logger.trace()) {
    logger.trace(
      "fruitninjascore:player %s destroyed fruits %s",
      playerId,
      stringify(gameObjectIds)
    );
  }
  let meta = this.meta;
  let scoreMap = meta.extras.points;
  let totalScore = 0;
  let fruits = game.gameObjectByIds(...gameObjectIds);
  let totalCuts = 0;
  if (logger.trace()) {
    logger.trace("fruitninjascore:total fruits %s", Object.keys(fruits).length);
  }
  for (let [id, fruit] of Object.entries(fruits)) {
    if (!fruit) continue;
    let score = 0;
    if (fruit.type === "bomb") {
      //get the players bombhit count
      let bombhitString = game.getPlayerProp(playerId, BOMB_HIT_KEY);
      let bombhit = parseInt(bombhitString);
      bombhit = isNaN(bombhit) ? 0 : bombhit;
      score = bombhit > 0 ? -30 : -20;
      game.setPlayerProp(playerId, BOMB_HIT_KEY, (bombhit + 1).toString());
    } else {
      score = scoreMap[fruit.type];
    }
    if (!score) continue;
    totalScore += score;
    totalCuts++;
  }
  logger.debug(
    "fruitninjascore:Total score for player %s is %s, cuts %s ",
    playerId,
    totalScore,
    totalCuts
  );
  //
  if (totalScore) {
    game.awardScore(playerId, { score: totalScore }, err => {
      if (err) {
        logger.error({ err: err });
        logger.error("neutrino core error , in awarding score");
      }
    });
    logger.debug(
      "fruitninjascore:score %s is awarded to player %s ",
      totalScore,
      playerId
    );
  }

  let combosizeachieved = 0;
  let combopoints = meta.extras.combopoints;
  if (combopoints) {
    let comboPointsMessages = meta.extras.combopointsmessages;
    totalCuts--;
    let combopointsIdx =
      totalCuts > combopoints.length - 1 ? combopoints.length - 1 : totalCuts;
    let comboPointsEarned = combopoints[combopointsIdx];
    logger.trace(
      "fruitninjascore: combopointsIdx %s, comboPointsEarned %s",
      combopointsIdx,
      comboPointsEarned
    );
    if (comboPointsEarned > 0) {
      combosizeachieved = combopointsIdx + 1;
      game.awardScore(
        playerId,
        {
          score: comboPointsEarned,
          message: JSON.stringify({
            count: totalCuts + 1,
            displayText: comboPointsMessages[combopointsIdx]
          })
        },
        err => {
          if (err) {
            logger.error({ err: err });
            logger.error("neutrino core error , in awarding score");
          }
        }
      );
    }
  }
  //look for combo of combos
  if (combosizeachieved) {
    checkComboOfCombos.call(this, game, combosizeachieved);
  }
};

const checkComboOfCombos = function(game, combosizeachieved, playerId) {
  let comboArr = game.getTransient(COMBO_ARR);
  if (!comboArr) {
    comboArr = [];
    game.addTransient(COMBO_ARR, comboArr);
    comboArr.push(combosizeachieved);
    return;
  }
  comboArr.push(combosizeachieved);

  for (let i = 0; i < this.comboofcombossorted; i++) {
    let comboofcombo = this.comboofcombossorted[i];
    if (comboOfComboSatisfied(comboArr, comboofcombo)) {
      //award the additional score
      game.awardScore(
        playerId,
        {
          score: comboofcombo,
          message: comboofcombo.winText
        },
        err => {
          if (err) {
            logger.error({ err: err });
            logger.error("neutrino core error , in awarding score");
          }
        }
      );

      break;
    }
  }
};

const comboOfComboSatisfied = function(combos, ccdef) {
  if (combos.length < ccdef.minCombosRequired) return false;
  for (let i = combos.length - 1; i >= 0; i--) {
    let combo = combos[i];
    if (combo < ccdef.minSize) return false;
  }
  return true;
};

const isKnife = function(id) {
  return id.startsWith(KNIFE_ID);
};
const collision = function(game, playerId, id1, id2) {
  if (!isKnife(id1) && !isKnife(id2)) return;
  let knifeId = null;
  let fruitId = null;
  if (isKnife(id1)) {
    knifeId = id1;
    fruitId = id2;
  } else {
    knifeId = id2;
    fruitId = id1;
  }
  let fruits = game.gameObjectByIds([fruitId]);
  if (!fruits) return;

  let fruit = fruits[fruitId];
  if (!fruit) return;
  if (fruit.type === "pomegranate") {
    game.awardScore(
      playerId,
      {
        score: 2
      },
      err => {
        if (err) {
          logger.error({ err: err });
          logger.error("neutrino core error , in awarding score");
        }
      }
    );
  }
};
module.exports = {
  playerDisconnection,
  playerJoined,
  gameObjectsDestroyed,
  collision
};
